import sys
import pickle
import json
import pandas as pd
import re

from jms.utils.git_download import * 
from jms.formula import *

from mass2chem.formula import *

def list_all_identifiers(cobra_model,notes):
    '''
    List all the identifier sources and the example of identifiers in the current model
    It is used to make strategy for parsing. 

    Input
    =====
    cobra_model: a cobra model
    notes: dependent on the identifier source, it may differ. In AGORA and EBI models, 'notes' are by default while in humanGEM, 'annotation' are by default

    Output
    ======
    A set of unique identifier source and examples of the identifiers

    '''
    identifiers = []
    example = {}
    for item in cobra_model.metabolites:
        try:
            identifiers.extend(item.__dict__[notes].keys())
            for k,v in item.__dict__[notes].items(): 
                if k not in example:
                    example.update({k:v})
        except:
            None
    uni_id = set(identifiers)
    return(uni_id,example)


def dict2listOfTuple(db_dict, delimiter):
    '''
    Convert dictionary to a list of tuple for convenient conversion in db_ids
    '''
    list_of_tuple = []
    for k,v in db_dict.items():
        if delimiter in v:
            list_of_ids = [x.rstrip() for x in v.split(delimiter)]
            for _id in list_of_ids:
                list_of_tuple.append((k,_id))
        else:
            list_of_tuple.append((k,v))
    return list_of_tuple

def listOfTuple2dict(tuples, delimiter = None):
    '''
    Convert a list of tuple to dictionary for convenient conversion in db_ids
    '''
    res_dict = {}
    unique_db_name = list(set([tup[0] for tup in tuples]))
    for db_name in unique_db_name:
        db_ids = []
        for tup in tuples:
            if tup[0] == db_name:
                db_ids.append(tup[1])
        if len(db_ids) == 1:
            db_ids = db_ids[0]
            res_dict[db_name] = db_ids
        elif len(db_ids) > 1:
            if delimiter == None:
                res_dict[db_name] = db_ids
            else:
                res_dict[db_name] = ', '.join(db_ids)
    return res_dict

def neutral_formula2mass(neutral_formula:str)->float:
    """Convert neutral formula to mass but removing characters (e.g., X, R) typical in GEM but not ready for metabolomics application

    Parameters
    ----------
    neutral_formula : str
        Neutral Formula

    Returns
    -------
    float
        Mass of the neutral formula
    Examples
    --------
    >>> from jms.utils.gems import *
    >>> neutral_formula2mass('C17H32O2')
    268.24023
     
    """
    formula_dict = parse_chemformula_dict(neutral_formula)
    if ("R" not in formula_dict) & ("X" not in formula_dict) & ("Z" not in formula_dict) & (len(formula_dict) != 0) :
        mono_mass = calculate_mass(formula_dict,6)
    else:
        mono_mass = float(0)
    return mono_mass

def remove_compartment_by_substr(Cpd_id:str,len_of_suffix:int)->str:
    """    
    remove compartment based the length of suffix of the compound ID.
    This is working especially for humanGEM as they just use the last character for compartmentalization designation
    However, this is not working for EBI model as they do have "_bm" alone with "_c"

    Parameters
    ----------
    Cpd_id : str
        Compound id from GEMs db
    len_of_suffix : int
        Length of suffix need to cut

    Returns
    -------
    str
        Compound id after remove

    Examples
    --------
    >>> remove_compartment_by_substr('MAM00001c', 1)
    'MAM00001'  

    """
    Cpd_id_mod = Cpd_id.rstrip()[:len(Cpd_id)-len_of_suffix]
    return Cpd_id_mod

def remove_compartment_by_split(Cpd_id,delimiter):
    '''
    remove compartment based the delimiters of the compound ID.
    This works for most other models.
    '''
    Cpd_id_mod = Cpd_id.rsplit(delimiter,1)[0]
    return(Cpd_id_mod)

def remove_duplicate_cpd(list_of_Cpds:list)->list:
    """remove duplicated compounds after removing the compartment information.

    Parameters
    ----------
    list_of_Cpds : list
        Raw compound list

    Returns
    -------
    list
        Compound list after deduplication
     
    """
    id_list = []
    new_list_of_Cpds = []
    for Cpd in list_of_Cpds:
        if Cpd.id not in id_list:
            id_list.append(Cpd.id)
            new_list_of_Cpds.append(Cpd)
        else:
            continue
    return new_list_of_Cpds

def remove_duplicate_rxn(list_of_Rxns:list)->list:
    """Removing uniport or exchange reactions.

    Parameters
    ----------
    list_of_Rxns : list
        Raw reaction list

    Returns
    -------
    list
        Reaction list after deduplication
     
    """
    new_list_of_Rxns = []
    for Rxn in list_of_Rxns:
        if Rxn.reactants == Rxn.products:
            continue
        else:
            new_list_of_Rxns.append(Rxn)
    return new_list_of_Rxns

def retain_valid_Rxns_in_Pathways(list_of_pathways:list, list_of_reactions:list)->list:
    """retain only valid Reactions in the pathway list after transport reactions excluded.

    Parameters
    ----------
    list_of_pathways : list
        list of pathways after port_pathway()
    list_of_reactions : list
        list of reactions after port_reaction()

    Returns
    -------
    list
        list of pathways after transport reactions excluded
     
    """
    valid_Rxns_list = [x.id for x in list_of_reactions]
    new_list_of_pathways = []
    for path in list_of_pathways:
        new_list_of_reactions = []
        for rxn in path.list_of_reactions:
            if rxn in valid_Rxns_list:
                new_list_of_reactions.append(rxn)
        path.list_of_reactions = new_list_of_reactions
        new_list_of_pathways.append(path)
    return new_list_of_pathways

def export_pickle(export_file_path:str,MetabolicModel):
    """Export metDataModel.core.MetabolicModel to pickle

    Parameters
    ----------
    export_file_path : str
        pickle path
    MetabolicModel : metDataModel.core.MetabolicModel
        metabolic model to export
     
    """
    with open(export_file_path, 'wb') as f:
        pickle.dump(MetabolicModel.serialize(), f, pickle.HIGHEST_PROTOCOL)

def export_json(export_file_path:str,MetabolicModel):
    """Export metDataModel.core.MetabolicModel to json

    Parameters
    ----------
    export_file_path : str
        json path
    MetabolicModel : metDataModel.core.MetabolicModel
        metabolic model to export
     
    """
    s = json.JSONEncoder().encode( MetabolicModel.serialize() )
    with open(export_file_path, 'w') as f:
        f.write(s)

def export_table(export_file_path:str,MetabolicModel,list_of_entries:str):
    """Export compounds, reactions and pathways in metabolic model into .csv

    Parameters
    ----------
    export_file_path : str
        csv file path
    MetabolicModel : metDataModel.core.MetabolicModel
        the model store metabolites, reactions and pathways
    list_of_entries : str
        suffix of .csv file
     
    """
    pd.DataFrame(MetabolicModel.serialize()[list_of_entries]).to_csv(export_file_path, index = False)

def fetch_MetabAtlas_GEM_identifiers(compound_list,
                                     modelName,
                                     local_path,
                                     metab_file_name = 'metabolites.tsv',
                                     overwrite = True):
    url_met = f'https://github.com/SysBioChalmers/{modelName}/blob/main/model/metabolites.tsv'
    git_download_from_file(url_met,local_path,metab_file_name)
    metab_df = pd.read_csv(os.path.join(local_path,metab_file_name),sep = '\t')
    metab_df.index = metab_df.mets
    metab_dict = metab_df.to_dict('index')

    metab_id_list = list(set([x[0:len(x)-1] for x in metab_dict.keys()]))
    decomp_metab_dict = {}
    for Id_decomp in metab_id_list:
        for Id_comp,v in metab_dict.items():
            if Id_decomp in Id_comp:
                decomp_metab_dict.update({Id_decomp:v})
                break
    for Cpd in compound_list:
        for k,v in decomp_metab_dict.items():
            new_db_ids = []
            if Cpd.id == k:
                if overwrite == True:
                    for kk,vv in v.items():
                        if re.match('met(.*)ID',kk) and isinstance(vv,str):
                            sub_key = re.match('met(.*)ID',kk)[1]
                            if ';' in vv:
                                sub_list = vv.split(';')
                                for sub_id in sub_list:
                                    new_db_ids.append((sub_key,sub_id))
                            else:
                                new_db_ids.append((sub_key,vv))
                    Cpd.db_ids = new_db_ids

def fetch_AGORA_GEM_identifiers(compound_list,
                                json_path,
                                overwrite = True):
    with open(json_path,'r') as f:
        list_vmh_cpd = json.load(f)
    vmh_dict = {}
    for vmh_cpd in list_vmh_cpd:
        vmh_dict.update({vmh_cpd['id']:vmh_cpd})
    new_cpd_list = []
    for myCpd in compound_list:
        for k,v in vmh_dict.items():
            if myCpd.id == k:
                if overwrite == True:
                    myCpd.db_ids = v['identifiers'] # the vmh json is using `identifiers` rather than `db_ids`
                break
        new_cpd_list.append(myCpd)
    return new_cpd_list


def fetch_CarveMe_GEM_charge_formula(compound_list,Bigg_json_path,overwrite = True):
    with open(Bigg_json_path,'r') as f:
        list_Bigg_cpd = json.load(f)
    Bigg_dict = {}
    for Bigg_cpd in list_Bigg_cpd:
        Bigg_dict.update({Bigg_cpd['id']:Bigg_cpd})
    new_cpd_list = []
    for myCpd in compound_list:
        for k,v in Bigg_dict.items():
            if myCpd.id == k:
                if overwrite == True:
                    myCpd.charge = v['charge']
                    myCpd.charged_formula = v['charged_formula'] # the Bigg json is using `identifiers` rather than `db_ids`
                break
        new_cpd_list.append(myCpd)
    return new_cpd_list


"""
Misc helper functions.

"""
import os
import shutil
import sys
import matchms
import scipy.stats
import numpy as np
import pandas as pd

from mass2chem.formula import calculate_formula_mass
from matchms.Spectrum import Spectrum
from .MSnSpectrum import MS2Spectrum

def flatten_nested_dicts(nested_dicts):
    """
    For nested dictionaries, i.e., where the value for a key in a dictionary is a dictionary
    return a flat, dictionary where the nested keys become top level keys. 

    If the same key is used at different levels, the one encountered last will be the 
    top level key, value pair.

    Args:
        nested_dicts (dict): a dictionary that has a dictionary as a value

    Returns:
        dict: flattened dict
    """    
    _d = {}
    for k, v in nested_dicts.items():
        if not isinstance(v, dict):
            _d[k] = v
        else:
            _d.update(flatten_nested_dicts(v))
    return _d


def extract_CD_csv(
    cd_csvs, ionization_mode, min_peaks=1, precursor_to_use="Confirmed", lazy=True
):
    """
    For a list of compound discover (CD) CSV libraries, read them into a form that is 
    usable for level1a annotation.

    In lazy mode, this yields the library entries, else a list of them is return.

    Args:
        CD_csv (list): list of paths to cd csv libraries
        ionization_mode (str): the ionization mode for the experiment (and library)
        min_peaks (int, optional): entries with fewer than these many peaks are discarded. Defaults to 1.
        precursor_to_use (str, optional): CD library has "confirmed" and "extracted" precursors, specifies which to use as the precursor m/z. Defaults to "Confirmed".
        lazy (bool, optional): determines if this is a generator or returns a list. Defaults to True.

    Returns:
        list: list of MS2 spectra

    Yields:
        MSnSpectrum: object representing an MSnSpectrum
    """
    standards_spectra = []
    for cd_csv in cd_csvs:
        for standard in pd.read_csv(cd_csv).to_dict(orient="records"):
            mzs, intensities = [], []
            for i in range(10):
                mzs.append(
                    standard["Confirm Extracted." + str(i)]
                    if i and "Confirm Extracted." + str(i) in standard
                    else standard["Confirm Extracted"]
                )
                intensities.append(
                    standard["Target Ratio." + str(i)]
                    if i and "Target Ratio." + str(i) in standard
                    else standard["Target Ratio"]
                )
            mzs = [mz for mz in mzs if (mz and not np.isnan(mz))]
            intensities = [i for i in intensities if (i and not np.isnan(i))]
            mzs, intensities = zip(*sorted(zip(mzs, intensities)))
            if len(mzs) == len(intensities) and len(mzs) >= min_peaks:
                spectrum = Spectrum(
                    mz=np.asarray([float(x) for x in mzs]),
                    intensities=np.asarray([float(x) for x in intensities]),
                    metadata=standard,
                )
                # assume proton for ionization adduct
                if ionization_mode == "pos":
                    standard["Theoretical_Precursor"] = (
                        calculate_formula_mass(standard["ChemicalFormula"])
                        + calculate_formula_mass("H")
                        - calculate_formula_mass("e")
                    )
                else:
                    standard["Theoretical_Precursor"] = (
                        calculate_formula_mass(standard["ChemicalFormula"])
                        - calculate_formula_mass("H")
                        + calculate_formula_mass("e")
                    )
                spectrum = process_ms2_spectrum(
                    {
                        "rt": standard["RT"] * 60,
                        "prec_mz": standard["Confirm Precursor"]
                        if precursor_to_use == "Confirmed"
                        else standard["Theoretical_Precursor"],
                        "cpd_name": standard["CompoundName"],
                        "spectrum": spectrum,
                    },
                    filename=cd_csv,
                    min_peaks=min_peaks,
                    skip_meta=True,
                    skip_filters=False,
                )
            if spectrum and not lazy:
                standards_spectra.append(spectrum)
            if spectrum and lazy:
                yield spectrum
    return standards_spectra


def get_parser(file_extension):
    """
    This will return the correct parser for a given MS2 file from
    matchms.

    :param file_extension: the extension of a file, e.g., .msp, .mzml

    :return: appropriate parser method from matchms
    """
    method = getattr(matchms.importing, "load_from_" + file_extension.lower())
    if method:
        return method
    print("No matching parser found for: ", file_extension)
    sys.exit()


def get_similarity_method(method_name):
    """
    This will return the specified similarity method from matchms.

    :param method_name: the name of a similarity method in matchms

    :return: appropriate similarity method from matchms
    """

    method = getattr(matchms.similarity, method_name)
    if method:
        return method
    print("No matching similarity method found for: ", method_name)
    sys.exit()


def lazy_extract_ms2_spectra(ms2_files, mz_tree=None):
    """
    This method takes a list of ms2 files and yields the MS2 spectrum 
    for each entry. 

    Args:
        ms2_files (list): list of ms2_files, can be any format matchms supports.
        mz_tree (interval_tree, optional): interval tree of precursor mzs, if provided, only 
        spectra whose precuror matches the tree will be returned. Defaults to None.

    Yields:
        MSnSpectrum: object representing an MSnSpectrum
    """
    for ms2_file in [ms2_files] if isinstance(ms2_files, str) else ms2_files:
        for spectrum in get_parser(ms2_file.split(".")[-1])(
            ms2_file, metadata_harmonization=True
        ):
            spectrum = matchms.filtering.add_precursor_mz(spectrum)
            if spectrum:
                precursor_mz = spectrum.get("precursor_mz", None)
            if precursor_mz:
                if mz_tree:
                    if mz_tree.at(precursor_mz):
                        ms2_spec_obj = process_ms2_spectrum(spectrum, filename=ms2_file)
                        if ms2_spec_obj:
                            yield ms2_spec_obj
                else:
                    ms2_spec_obj = process_ms2_spectrum(spectrum, filename=ms2_file)
                    if ms2_spec_obj:
                        yield ms2_spec_obj


def process_ms2_spectrum(
    spectrum, filename="not_specified", min_peaks=1, skip_meta=False, skip_filters=False
):
    """
    This is the default MS2 processing used by the pipeline.

    This adds the precursor_mz, applies the default filters, normalizes
    the intensiies requires a minimum number of peaks and then converts
    the retention time to seconds.

    :param filename: used to record what file the spectrum came from
    :param min_peaks: the min number of peaks a spectrum must have

    :return: dictionary summarize MS2 spectrum.
    """

    if not skip_meta:
        spectrum = matchms.filtering.add_precursor_mz(spectrum)
        if not skip_filters:
            spectrum = matchms.filtering.default_filters(spectrum)
            spectrum = matchms.filtering.normalize_intensities(spectrum)
            spectrum = matchms.filtering.require_minimum_number_of_peaks(
                spectrum, min_peaks
            )

        if spectrum is None:
            return None
        try:
            spectrum.set("retention_time", spectrum.metadata["scan_start_time"][0] * 60)
        except:
            pass

        spec_id = []
        if spectrum.get("precursor_mz"):
            spec_id.append(str(float(spectrum.get("precursor_mz", ""))))
        if spectrum.get("retention_time"):
            spec_id.append(str(float(spectrum.get("retention_time", ""))))
        spec_id.append(filename)
        spec_id = "_".join(spec_id)
        reference_id = spectrum.get("compound_name", spec_id)
        spectrum = MS2Spectrum(
            spec_id=spec_id,
            precursor_mz=float(spectrum.get("precursor_mz")),
            precursor_rt=float(spectrum.get("retention_time")) if spectrum.get("retention_time") else None,
            list_mz=[],
            list_intensity=[],
            matchms_spectrum=spectrum,
            source=filename,
            instrument="Unknown",
            collision_energy="Unknown",
            compound_name=reference_id,
        )
    else:
        spec_id = []
        spec_id.append(str(float(spectrum["prec_mz"])))
        spec_id.append(str(float(spectrum["rt"])))
        spec_id.append(filename)
        spec_id = "_".join(spec_id)
        precursor_mz = spectrum["prec_mz"]
        precursor_rt = spectrum["rt"]
        reference_id = spectrum["cpd_name"]
        spectrum = spectrum["spectrum"]
        if not skip_filters:
            spectrum = matchms.filtering.default_filters(spectrum)
            spectrum = matchms.filtering.normalize_intensities(spectrum)
            spectrum = matchms.filtering.require_minimum_number_of_peaks(
                spectrum, min_peaks
            )
        spectrum.set("retention_time", precursor_rt)
        spectrum.set("precursor_mz", precursor_mz)

        spectrum = MS2Spectrum(
            spec_id=spec_id,
            precursor_mz=float(precursor_mz),
            precursor_rt=float(precursor_rt),
            list_mz=[],
            list_intensity=[],
            matchms_spectrum=spectrum,
            source=filename,
            instrument="Unknown",
            collision_energy="Unknown",
            compound_name=reference_id,
        )

    return spectrum


def search_for_mzml(sdir):
    """
    Given a directory, encrively search for all files with an .mzml
    extension and return their paths.

    :param sdir: the directory in which to search

    :return: list of absolute mzML filepaths in sdir
    """
    mzml_found = []
    for d, _, fs in os.walk(sdir):
        for f in [f for f in fs if f.endswith(".mzML")]:
            mzml_found.append(os.path.join(os.path.abspath(d), f))
    return mzml_found


def recursive_encoder(to_encode):
    """
    This method takes a datastructure and makes it JSON-ready by
    recursively calling itself on the datastructure at all depths.
    This is needed because some objects such as Acquisition have their own
    method called JSON_repr that will encode it correctly whereas the
    JSON library will not.

    Thus this can encode any datastructure comprised of other objeccs
    provided that they too are either a primitive type (e.g., str, int)
    or implement their own JSON_repr method.

    :param to_encode: datastructure for conversion to JSON

    :return: JSON-encodable representation of to_encode
    """
    if isinstance(to_encode, (str, int, float, type(None))):
        return to_encode
    elif isinstance(to_encode, (list, set)):
        return [recursive_encoder(x) for x in to_encode]
    elif isinstance(to_encode, dict):
        return {k: recursive_encoder(v) for k, v in to_encode.items()}
    elif hasattr(to_encode, "json_repr"):
        return to_encode.json_repr
    elif hasattr(to_encode, "serialize"):
        return recursive_encoder(to_encode.serialize())


# the following define mappings between strings and functions used in
# many places in the pcpfm. New modes can be added trivially by adding
# them to these mappings.

correlation_modes = {
    "spearman": scipy.stats.spearmanr,
    "kendall": scipy.stats.kendalltau,
    "pearson": np.corrcoef,
}

descriptive_stat_modes = {
    "median": np.median,
    "max": np.max,
    "min": np.min,
    "mean": np.mean,
}

log_modes = {
    "log2": np.log2, 
    "log10": np.log10
    }

def delete_dir_or_file(to_delete):
    try:
        shutil.rmtree(to_delete)
    except:
        try:
            os.remove(to_delete)
        except:
            "Unable to delete: " + to_delete

file_operations = {
    "link": os.link,
    "copy": shutil.copy,
    "move": shutil.move,
    "delete": delete_dir_or_file
}



"""
MS2Spectrum contains the constructors for MS2 spectra. These are used for L1a and L2 annotations. 

The actual MS2 comparisons are performed by matchms in the EmpCpds object.

"""

import os
from matchms.Spectrum import Spectrum
import numpy as np
from metDataModel import core

class MS2Spectrum(core.Spectrum):
    """
    This represents an MS2 spectrum
    """
    def __init__(
        self,
        spec_id,
        precursor_mz,
        precursor_rt,
        list_mz=None,
        list_intensity=None,
        matchms_spectrum=None,
        source="",
        instrument=None,
        collision_energy=None,
        compound_name=None,
        annotations=None,
    ):
        """_summary_

        Args:
            id (str): a string that uniquely identifies the MS2 spectrum
            precursor_mz (float): the m/z of the precursor ion
            precursor_rt (float): the retention time for the precurso ion
            list_mz (list, optional): list of m/z values for the spectrum. Defaults to None.
            list_intensity (list, optional): list of intensities for the spectrum co-indexed with list_mz. Defaults to None.
            matchms_spectrum (object, optional): a matchms spectrum object. Defaults to None.
            source (str, optional): the file from which the ms2 spectrum was created. Defaults to "".
            instrument (str, optional): the type of instrument from which the spectrum was acquired. Defaults to None.
            collision_energy (float, optional): the collision energey in eV. Defaults to None.
            compound_name (str, optional): the name of the compound. Defaults to None.
            annotations (list, optional): list of annotations for the spectrum. Defaults to None.
        """
        super().__init__(spec_id)
        source = os.path.basename(source) if source != "" else os.path.basename(source)
        self.precursor_ion_id = (str(precursor_mz) + "_" + str(precursor_rt) + "_" + os.path.basename(source))
        self.spec_id = spec_id
        self.rtime = precursor_rt
        self.retention_time = precursor_rt
        self.precursor_ion_mz = precursor_mz
        self.instrument = instrument
        self.collision_energy = collision_energy
        self.matchms_spectrum = matchms_spectrum
        if self.matchms_spectrum:
            self.list_mz = [x[0] for x in self.matchms_spectrum.peaks]
            self.list_intensity = [x[1] for x in self.matchms_spectrum.peaks]
            self.min_mz = min(self.list_mz)
            self.max_mz = max(self.list_mz)
        elif list_mz and list_intensity:
            self.list_mz = list_mz
            self.list_intensity = list_intensity
        else:
            self.list_mz = []
            self.list_intensity = []


        self.annotations = [] if annotations is None else annotations
        self.compound_name = compound_name
        self.source = source

    @property
    def prec_mz(self):
        """
        Simply a shortcut for accessing the precursor_ion_mz field. This is used because the 
        precursor_ion_mz is long and can yield lines that are too long for pylint.

        Returns:
            float: precursor ion mz
        """
        return self.precursor_ion_mz

    @staticmethod
    def from_embedding(embedding):
        """
        This recreates the MS2 object from the serialized version of the object

        Args:
            embedding (dict): the serialized version of the MSnSpectrum.

        Returns:
            MS2Spectrum: the MSnSpectrum that was serialized
        """
        return MS2Spectrum(
            spec_id=None,
            precursor_mz=float(embedding["precursor_ion_mz"]),
            precursor_rt=float(embedding["rtime"]),
            list_mz=embedding["list_mz"],
            list_intensity=embedding["list_intensity"],
            matchms_spectrum=Spectrum(
                mz=np.array(list(embedding["list_mz"])),
                intensities=np.array(embedding["list_intensity"]),
                metadata={},
            ),
            source=embedding["source"],
            instrument=embedding["instrument"],
            collision_energy=embedding["collision_energy"],
            compound_name=None,
            annotations=embedding["annotations"],
        )

    def annotate(self, other_ms2, score, matched_peaks, annotation_level="Unspecified"):
        """
        Given another MSnSpectrum object, annotate this MSnSpectrum.

        Args:
            other_MS2 (object): the other MS2 spectrum, typically from a database
            score (float): the score from the similarity comparison
            matched_peaks (int): number of peaks 
            annotation_level (str, optional): a string designating the annotation level. Defaults to "Unspecified".
        """
        self.annotations.append(
            {
                "msms_score": score,
                "matched_peaks": matched_peaks,
                "db_precursor_mz": other_ms2.precursor_ion_mz,
                "db_precursor_rt": other_ms2.rtime,
                "reference_id": other_ms2.compound_name,
                "list_mz": [x[0] for x in other_ms2.matchms_spectrum.peaks],
                "list_intensity": [x[1] for x in other_ms2.matchms_spectrum.peaks],
                "primary_db": other_ms2.source,
                "source": self.source,
                "annotation_level": annotation_level,
            }
        )

    def embedding(self):
        """
        Given an MSnSpectrum, generate a serializable version of the object.

        Returns:
            dict: serialized version
        """
        embedding = {}
        for k, v in self.__dict__.items():
            if isinstance(v, (int, float, str, dict, set, list)):
                embedding[k] = v
        return embedding



'''
Functions related to isotopes and adductions.
Using mass2chem.

compute_adducts_formulae(304.2038, 'C19H28O3',  mode='pos', primary_only=True)
Out[5]:
[[304.203251, 'M[1+]', 'C19H28O3'],
 [305.21107646677, 'M+H[1+]', 'C19H29O3'],
 [327.19307646677, 'M+Na[1+]', 'C19H28NaO3'],
 [323.22167646677, 'M+H2O+H[1+]', 'C19H31O4']]

'''

from khipu.utils import adduct_search_patterns, \
                            adduct_search_patterns_neg, \
                                isotope_search_patterns, \
                                    extended_adducts

from mass2chem.formula import compute_adducts_formulae
from mass2chem.formula import parse_chemformula_dict
import json
import heapq
from scipy.stats import multinomial
import numpy as np
import itertools
from copy import deepcopy

dev_isotopic_patterns = [
    # mass diff, isotopes, (intensity ratio constraint)
    (1.003355, '13C/12C', (0, 0.8)),      # 13C-12C, 12C~99%, 13C ~ 1%
    (0.997035, '15N/14N', (0, 0.2)),     # 15N-14N, 14N ~ 99.64%, 15N ~ 0.36%
    (2.004245, '18O/16O', (0, 0.2)),      # 18O-16O, 16O ~ 99.76, 16O ~ 0.2%
    (1.995796, '34S/32S', (0, 0.4)),      # 32S (95.02%), 33S (0.75%), 34S (4.21%)
    (0.999388, '33S/32S', (0, 0.1)),
    # double isotopes
    (2.00039, 'M(13C),M(15N)', (0, 0.2)),
    (2.999151, 'M(13C),M(34S)', (0, 0.4)),
    # double charged
    (0.5017, '13C/12C, double charged', (0, 0.8)),
    (0.4985, '15N/14N, double charged', (0, 0.2)),
]

#
# -----------------------------------------------------------------------------
#

isotope_information = json.load(open("/Users/mitchjo/Projects/xasari/NIST_isotope_data.json"))

element_to_iso = {}
for iso_dict in isotope_information:
    if iso_dict["Atomic Symbol"] == "D":
        iso_dict["Atomic Symbol"] = "H"
    if iso_dict["Atomic Symbol"] not in element_to_iso:
        element_to_iso[iso_dict["Atomic Symbol"]] = []
    element_to_iso[iso_dict["Atomic Symbol"]].append(iso_dict)

element_to_iso_tuples = {}
for element, isotopes in element_to_iso.items():
    iso_NAPs, iso_masses = [], []
    for iso in isotopes:
        iso_symbol = iso["Atomic Symbol"] + str(iso["Mass Number"])
        mass = iso["Relative Atomic Mass"]
        NAP = iso['Isotopic Composition']
        if NAP and NAP > .005:
            iso_NAPs.append((iso_symbol, NAP, mass))
    iso_NAPs = sorted(iso_NAPs, key=lambda x: -x[1])
    if iso["Atomic Symbol"] == "Sn" or iso["Atomic Symbol"] == "Os":
        iso_NAPs = iso_NAPs[:2]
    if iso_NAPs:
        element_to_iso_tuples[element] = iso_NAPs

multi_cache = {}
element_prob_vectors = {}
def calc_prob(iso_coord):
    sub_coord_probs = []
    for element, sub_coord in iso_coord.items():
        key_tuple = (element, tuple(sub_coord))
        if key_tuple in multi_cache:
            sub_coord_probs.append(multi_cache[key_tuple])
        else:
            key_sum = (element, sum(sub_coord))
            if element not in element_prob_vectors:
                element_prob_vectors[element] = np.array([x[1] for x in element_to_iso_tuples[element]])
            if key_sum not in multi_cache:
                multi_cache[key_sum] = multinomial(key_sum[1], element_prob_vectors[element])
            multi_cache[key_tuple] = multi_cache[key_sum].pmf(sub_coord)
            sub_coord_probs.append(multi_cache[key_tuple])
    return np.product(sub_coord_probs)

mass_cache = {}
element_mass_vectors = {}
def calc_mass(iso_coord):
    sub_coord_masses = []
    for element, sub_coord in iso_coord.items():
        key_tuple = (element, tuple(sub_coord))
        if key_tuple in mass_cache:
            sub_coord_masses.append(mass_cache[key_tuple])
        else:
            if element not in element_mass_vectors:
                element_mass_vectors[element] = np.array([x[2] for x in element_to_iso_tuples[element]])
            mass_cache[key_tuple] = np.dot(sub_coord, element_mass_vectors[element])
        sub_coord_masses.append(mass_cache[key_tuple])
    return np.sum(sub_coord_masses)

neighbor_cache = {}
delta_cache = {}
def gen_neighbors(iso_coord):
    new_coords = []
    for _, sub_coord in iso_coord.items():
        key = tuple(sub_coord)
        if key not in neighbor_cache: 
            neighbor_cache[key] = [sub_coord]
            if len(sub_coord) > 1:
                if len(sub_coord) not in delta_cache:
                    delta_cache[len(sub_coord)] = np.array([x for x in itertools.permutations([-1, 1] + [0 for _ in range(len(sub_coord)-2)])], dtype=np.int16)           
                for delta in delta_cache[len(sub_coord)]:
                    new_coord = delta + sub_coord
                    if not np.any(new_coord < 0):
                        neighbor_cache[key].append(new_coord)
        new_coords.append(neighbor_cache[key])
    new_iso_coords = []
    for coord_mask in np.ndindex(tuple([len(x) for x in new_coords])):
        if not np.all(coord_mask == 0):
            new_iso_coord = {}
            for i, (index, k) in enumerate(zip(coord_mask, iso_coord.keys())):
                new_iso_coord[k] = new_coords[i][index]
            new_iso_coords.append(new_iso_coord)
    return new_iso_coords

def hash_isotopologue(isotopologue):
    return "|".join([k + ','.join([str(x) for x in v]) for k,v in isotopologue.items()])

delta_substrings = {}
isotopes_vectors = {}
def delta_info(isotopologue, reference):
    delta_string = []
    for element in isotopologue.keys():
        iso_sub_coord, ref_sub_coord = isotopologue[element], reference[element]
        key = (element, tuple(iso_sub_coord), tuple(ref_sub_coord))
        if key not in delta_substrings:
            substring = []
            delta_values = iso_sub_coord - ref_sub_coord
            if element not in isotopes_vectors:
                isotopes_vectors[element] = [x[0] for x in element_to_iso_tuples[element]]
            for delta, iso_name in zip(delta_values, isotopes_vectors[element]):
                if delta > 1:
                    substring.append(str(delta) + iso_name)
                elif delta == 1:
                    substring.append(iso_name)
                else:
                    pass
                delta_substrings[key] = ",".join(substring)
        delta_string.append(delta_substrings[key])
    delta_string = '(' + ",".join([x for x in delta_string if x]) + ')'
    if delta_string != '()':
        return delta_string, calc_mass(isotopologue) - calc_mass(reference)
    else:
        return '', calc_mass(isotopologue) - calc_mass(reference)


def gen_isotopologues(formula, NAP_cutoff=0):
    iso_coord = {}
    for k, v in parse_chemformula_dict(formula).items():
        iso_coord[k] = [0 for _ in element_to_iso_tuples[k]]
        iso_coord[k][0] = v
        iso_coord[k] = np.array(iso_coord[k], dtype=np.int16)
    iso_heap = []
    isos = [(-1 * calc_prob(x), x) for x in [iso_coord] + gen_neighbors(iso_coord)]
    isos = sorted(isos, key=lambda x: x[0])
    reference = isos[0][1]
    used = set()
    for NAP_prob, iso_coord in isos:
        if -1 * NAP_prob > NAP_cutoff:
            iso_hash = hash_isotopologue(iso_coord)
            if iso_hash not in used:
                heapq.heappush(iso_heap, (NAP_prob, iso_coord))
                used.add(hash_isotopologue(iso_coord))
    while iso_heap:
        NAP, most_abundant = heapq.heappop(iso_heap)
        yield -1 * NAP, most_abundant, calc_mass(most_abundant), delta_info(most_abundant, reference)
        for x in gen_neighbors(most_abundant):
            hash = hash_isotopologue(x)
            if hash not in used:
                NAP_prob = calc_prob(x)
                if NAP_prob > NAP_cutoff:
                    heapq.heappush(iso_heap, (-1 * NAP_prob, x))
                    used.add(hash)

def generate_ion_signature2(mw, neutral_formula, mode='pos', primary_only=True):
    adducts = compute_adducts_formulae(mw, neutral_formula, mode, primary_only)
    C13 = []
    for A in adducts:
        C13.append([
            A[0]+1.003355, A[1]+',C13', A[2]+',(C13)'
        ])
    return adducts + C13

def generate_ion_signature(mw, neutral_formula, mode='pos', primary_only=True, C13_only=False, NAP_cutoff=.01):
    '''
    Extend mass2chem.formula.compute_adducts_formulae by C13 or any number of isotopologues
    based on NAP cutoff.

    Note - Resulting chemical formula is not computable.
    '''
    adducts = compute_adducts_formulae(mw, neutral_formula, mode, primary_only)
    if C13_only:
        C13 = []
        for A in adducts:
            A.append(0)
            C13.append([
                A[0]+1.003355, A[1]+',C13', A[2]+',(C13)', 1
            ])
        return adducts + C13
    else:
        isotopologues = []
        for i, (NAP, _, _, (delta_string, delta_mass)) in enumerate(gen_isotopologues(neutral_formula)):
            if NAP > NAP_cutoff:
                for A in adducts:
                    if delta_string:
                        isotopologues.append([
                            A[0] + delta_mass,
                            A[1] + "," + delta_string[1:-1] + ";" + str(i),
                            A[2] + "," + delta_string, 
                            i,
                        ])
                    else:
                        isotopologues.append([
                            A[0] + delta_mass,
                            A[1] + ";" + str(i),
                            A[2],
                            i,
                        ])
            else:
                return isotopologues
        isotopologues = []
        for A in adducts:
            isotopologues.append([
                A[0],
                A[1] + ";0",
                A[2],
                0,
            ])
        return isotopologues



'''
Class structures to connect metabolite/compound databases to empirical compounds;
and to connect experimental data to empirical compounds.
'''

import json
from operator import itemgetter
import numpy as np
import tqdm
from khipu.epdsConstructor import epdsConstructor

from khipu.utils import adduct_search_patterns, \
                            adduct_search_patterns_neg, \
                                isotope_search_patterns, \
                                    extended_adducts

from .search import *
from .formula import *
from .ions import compute_adducts_formulae, generate_ion_signature
from .data.list_formula_mass import list_formula_mass


def annotate_peaks_against_kcds(list_peaks, 
                                list_compounds, 
                                export_file_name_prefix='jms_annotated_',
                                mode='pos',  
                                mz_tolerance_ppm=5, 
                                rt_tolerance = 2,
                                check_isotope_ratio = False):
    '''
    Wrapper function as example, to generate three annotation files for input list_peaks.
    list_compounds is known compound database, e.g.
    list_compounds = json.load(open('jms/data/compounds/list_compounds_HMDB4.json'))
    list_peaks is experimental data, list of features in json.
    '''
    KCD = knownCompoundDatabase()
    KCD.mass_index_list_compounds(list_compounds)
    KCD.build_emp_cpds_index()
    # optional to export_mass_indexed_compounds
    # KCD.export_mass_indexed_compounds(export_file_name_prefix+"KCD_mass_indexed_compounds.json")

    EED = ExperimentalEcpdDatabase(mode=mode, 
                                   mz_tolerance_ppm=mz_tolerance_ppm,
                                   rt_tolerance=rt_tolerance)
    EED.build_from_list_peaks(list_peaks)
    # Second, singletons that get a formula match in KCD
    EED.annotate_singletons(KCD)       
    # Third, the remaining features unmatched to anything (orphans). Exported for potential downstream work.
    # EED.dict_empCpds = self.append_orphans_to_epmCpds(EED.dict_empCpds)
    
    EED.export_annotations(KCD, export_file_name_prefix)


#----------------------------------------------------------------------------------------
class knownCompoundDatabase:
    '''
    Indexed data store for known compounds, e.g., from a database or a metabolic model. 
    The search should be on empCpd not Cpd.
    An empCpd is an empirical compound, a set of features drived from the same mass (see README), 
    which often include isomers. The regular mass search cannot distinguish isomers.

    One can search by mass or mass tree (patterns of isotopes/adducts, in the form of empCpd.
    We use build_centurion_tree to index empCpds;
    centurion_mass_tree is an indexed dictionary to group ions by 100th decimal.
    There are situations where only neutral mass is searched; others requiring ionized forms.
    This class generates three trees to accommodate them: 'neutral', 'pos' and 'neg'.
    '''
    def __init__(self):
        '''
        Two main data containers, mass_indexed_compounds and emp_cpds_trees.
        The latter is indexed for searches, separately for positive and negative ion modes.
        '''
        self.mass_indexed_compounds = {}
        self.emp_cpds_trees = { 'pos': {}, 
                                'neg': {},
                                'neutral': {},
                                }
        
    def mass_index_list_compounds(self, list_compounds):
        '''
        Using dictionaries to index data, then compile into list_emp_cpds.
        Input
        =====
        list of compounds:  Example of a compound
        {
            'primary_id': HMDB0000195,
            'primary_db': 'HMDB',
            'name': 'Inosine',
            "neutral_formula": C10H12N4O5,
            "neutral_formula_mass": 268.08077, 
            'SMILES': 'OC[C@H]1O[C@H]([C@H](O)[C@@H]1O)N1C=NC2=C(O)N=CN=C12', 
            'inchikey': 'UGQMRVRMYYASKQ-KQYNXXCUSA-N',
            'LogP': -2.10,
            'other_ids': {'PubChem': '6021',
                        'KEGG': 'C00294',
                        'ChEBI': '17596',
                        'MetaNetX': 'MNXM1103335',
                        },
        }
        See example of how to generate Compounds:
        https://github.com/shuzhao-li/JMS/wiki/Scripting-to-create-a-dictionary-from-HMDB-data

        Update
        ======
        self.mass_indexed_compounds so that isomers are grouped under same empCpd.
        '''
        from collections import defaultdict
        _db = defaultdict(list)
        for cpd in list_compounds:
            _db[cpd['neutral_formula'] + "_" + str(round(float(cpd['neutral_formula_mass']),6))].append(cpd)
        for k, v in _db.items():
            self.mass_indexed_compounds[k] = {
                "interim_id": k,
                "neutral_formula": v[0]['neutral_formula'],
                "neutral_formula_mass": float(v[0]['neutral_formula_mass']),
                "compounds": v
            }

    def build_emp_cpds_index(self, primary_only=True, include_C13=True):
        '''
        For each emp_cpd, generate ion signatures common_adducts adducts (pos or neg ion mode).
        Then use build_centurion_tree function from .search to build index.

        primary_only: only considering the most common ions 
        (https://github.com/shuzhao-li/mass2chem/blob/master/mass2chem/formula.py),
        which are adequate for initial search and annoation. 
        One may choose to extend the search for other ions after the emp_cpd is matched.

        include_C13: considering 13C isotopes for the adducts.
        This is not desired in default application of matching to experimental data via empCpds.
        The isotopes and adducts should be organized into empCpds in experimental data,
        then the 12C anchor ion has to be present in search result in oder to be considered a true match.
        In that scenario, adding 13C in DB records is unecessary.
        One should include 13C when generating a database for single ion searches.

        Format example -
        pos_peak_list: [{'parent_epd_id': 1670, 'mz': 133.0970237, 'ion_relation': 'M[1+]'}, ...]

        '''
        __ion_generator__ = compute_adducts_formulae
        if include_C13:
            __ion_generator__ = generate_ion_signature
        
        peak_lists = {"pos": [], "neg": [], "neutral": []}
        for k, v in tqdm.tqdm(self.mass_indexed_compounds.items()):
            peak = {a: b for a, b in v.items()}
            peak['mz'] = v['neutral_formula_mass']
            peak['parent_epd_id'] = k
            peak['ion_relation'] = 'neutral'
            peak_lists['neutral'].append(peak)
            for mode in ["pos", "neg"]:
                for ion in __ion_generator__(v['neutral_formula_mass'], v['neutral_formula'], mode=mode, primary_only=primary_only):
                    ion_peak = dict(peak)
                    ion_peak['mz'] = ion[0]
                    ion_peak['ion_relation'] = ion[1]
                    if "order" in ion:
                        ion_peak["order"]: ion[3]
                    peak_lists[mode].append(ion_peak)
        self.emp_cpds_trees = {k: build_centurion_tree(v) for k, v in peak_lists.items()}

    def search_mz_single(self, query_mz, mode='pos', mz_tolerance_ppm=5):
        '''
        return list of matched empCpds, e.g.
            [{'mz': 130.017306555, 'parent_epd_id': 'C4H3FN2O2_130.017856', 'ion_relation': 'M[1+]'}]
        '''
        return find_all_matches_centurion_indexed_list(query_mz, self.emp_cpds_trees[mode], mz_tolerance_ppm)

    def search_mz_batch(self, query_mz_list, mode='pos', mz_tolerance_ppm=5):
        results = []
        for query_mz in query_mz_list:
            results.append(
                self.search_mz_single(query_mz, mode, mz_tolerance_ppm)
            )
        return results

    def evaluate_mass_accuracy_ratio(self, query_mz_list, mode='pos', mz_tolerance_ppm=10):
        '''
        Find the best matches of query_mz_list in known compound database, within mz_tolerance_ppm,
        then evaluate systematic shift of experimental m/z values by theoretical compound values.
        This can be used to report mass accuracy, and to help mass calibration.
        '''
        results = []        # delta m/z, [(expt - theoretical)/theoretical, ...]
        for query_mz in query_mz_list:
            found = find_best_match_centurion_indexed_list(query_mz, self.emp_cpds_trees[mode], mz_tolerance_ppm)
            if found:
                results.append( query_mz/found['mz'] - 1 )

        if results:
            ratio = np.mean(results)
            print("Mass accuracy was estimated on %d matched values as %2.1f ppm." %(len(results), ratio*1000000))
            return ratio
        else:
            print("No m/z match found in database to estimate mass accuracy.")
            return None

    def search_emp_cpd_single(self, emp_cpd, mode='pos', mz_tolerance_ppm=5):
        '''
        emp_cpd : empirical compound, following loosely specificiations in metDataModel, e.g.
        {interim_id': 12,
        'neutral_formula_mass': None,
        'neutral_formula': None,
        'MS1_pseudo_Spectra': [{'id_number': 'F94',
                                'mz': 98.97531935309317,
                                'rtime': 700.0,
                                'charged_formula': '',
                                'ion_relation': 'anchor',
                                'parent_epd_id': 12}, ], ...}

        If input emp_cpd is generated from khipu, neutral_formula_mass is usually alrady assigned, 
        and this is simple search on neutral_formula_mass.
        Otherwise, anchor ion is searched in emp_cpds_trees (from mass_indexed_compounds).
        The ion_relations are expected to be ['M[1+]', 'M+H[1+]', 'M+Na[1+]', 'M+H2O+H[1+]'] for pos,
        and ['M[-]', 'M-H[-]', 'M-H2O-H[-]', 'M+Cl[-]'] for neg.

        This searches only neutral mass or anchor ion, because if anchor ion is not found in the database, 
        other adducts are not possible.

        Return
        ======
        list of matched db empCpd as [(empCpd_id, ion_relation, score), ...], sorted by score.
        While most search will return a single match. Occasionally, undeterministic situations appear,
        e.g. C6H13N and C6H14N could match to M+ and M+H+, respectively.
        The resolution of these will be left to downstream methods (bayesian etc).
        The score is number of matched ions, which may or may not be useful.
        '''
        results = []
        if emp_cpd['neutral_formula_mass']:
            matches = self.search_mz_single(emp_cpd['neutral_formula_mass'], 'neutral', mz_tolerance_ppm)
            for _M in matches:
                results.append((_M['parent_epd_id'],
                                _M['ion_relation'], 1))
        else:
            query_mzs = [x['mz'] for x in emp_cpd['MS1_pseudo_Spectra']]
            # look for anchor ion first; default first ion in 'MS1_pseudo_Spectra'
            matches = self.search_mz_single(query_mzs[0], mode, mz_tolerance_ppm)
            # format - [{'mz': 130.017306555, 'parent_epd_id': 'C4H3FN2O2_130.017856', 'ion_relation': 'M[1+]'}, ...]
            for _M in matches:
                R = self.mass_indexed_compounds[_M['parent_epd_id']]
                # compute_adducts_formulae considers both isotopes and adducts
                db_peaks = compute_adducts_formulae(R['neutral_formula_mass'], R['neutral_formula'], mode)
                results.append((_M['parent_epd_id'],
                                _M['ion_relation'],     # ion_relation btw anchor and DB record, not ions in empCpd
                                score_emp_cpd_matches(query_mzs, [x[0] for x in db_peaks], mz_tolerance_ppm), 
                        ))
        return sorted(results, key=itemgetter(2), reverse=True)
        
    def search_emp_cpd_batch(self, list_emp_cpd, mode='pos', mz_tolerance_ppm=5):
        results = []
        for emp_cpd in list_emp_cpd:
            results.append(
                self.search_emp_cpd_single(emp_cpd, mode, mz_tolerance_ppm)
            )
        return results

    def short_report_emp_cpd(self, interim_id):
        return ";".join([ C['primary_id']+'$'+C['name'] 
                        for C in self.mass_indexed_compounds[interim_id]['compounds'] ])

    def export_mass_indexed_compounds(self, outfile="KCD_mass_indexed_compounds.json"):
        with open(outfile, 'w', encoding='utf-8') as O:
            json.dump(self.mass_indexed_compounds, O, ensure_ascii=False, indent=2)

    def export_search_emp_cpd_batch(self, batch_result):
        '''
        '''
        pass


#----------------------------------------------------------------------------------------
class ExperimentalEcpdDatabase:
    '''
    Build a boutique data store for user's experimental data, list of JSON features/peaks.
    Allow search of Compounds, which support targeted compound search in a dataset.
    '''
    def __init__(self, 
                 mode='pos', 
                 mz_tolerance_ppm=5, 
                 rt_tolerance=2,
                 ):
        '''
        mode: ionizaation mode, 'pos' or 'neg'.
        mz_tolerance_ppm: ppm tolerance in examining m/z patterns.
        rt_tolerance: tolerance threshold for deviation in retetion time, arbitrary unit depending on input data.
                Default intended as 2 seconds.
        The isotope/adduct patterns here are populated via self.get_isotope_adduct_patterns, as imported from khipu.
        But one can directly overwrite the patterns for customization (e.g. in asari).
        '''
        self.mode = mode
        self.mz_tolerance_ppm = mz_tolerance_ppm
        self.rt_tolerance = rt_tolerance
        self.get_isotope_adduct_patterns()
        self.list_peaks = []
        self.indexed_peaks = {}
        self.dict_peaks = {}
        self.dict_empCpds = {}
        self.indexed_empCpds = {}
        self.peak_to_empCpd = {}
        self.peak_to_empCpd_ion_relation = {}

    def get_isotope_adduct_patterns(self, 
                                    adduct_search_patterns=adduct_search_patterns,
                                    adduct_search_patterns_neg=adduct_search_patterns_neg,
                                    isotope_search_patterns=isotope_search_patterns[:2],
                                    extended_adducts=extended_adducts
                                    ):
        '''
        Populate isotope/adduct patterns for this instance.
        The default isotope/adduct patterns here are imported from khipu.
        The default isotope_search_patterns are limited to M0 and M1.
        '''
        self.adduct_patterns = adduct_search_patterns
        if self.mode == 'neg':
            self.adduct_patterns = adduct_search_patterns_neg
        self.isotope_search_patterns = isotope_search_patterns
        self.extended_adducts = extended_adducts

    def build_from_list_peaks(self, 
                              list_peaks, 
                              charges=[1, 2, 3],
                              has_parent_masstrack=True,
                              ):
        '''
        Wrapper of khipu epdsConstructor, to construct empirical compounds (empCpds).
        Updates self.dict_empCpds and does self.index_empCpds().

        list_peaks : [{'parent_masstrace_id': 1670, 'mz': 133.09702315984987, 'rtime': 654, 
                'height': 14388.0, 'id_number': 555}, ...]
        isotope/adduct patterns, mz_tolerance_ppm and rt_tolerance are from self,
        as they are parameters to this class.

        charges : multiple charged ions to consider.
        has_parent_masstrack : True if from asari. 
            Used to determine the need of setting parent_masstrack.

        Updates
        -------
        self.list_peaks
        self.dict_empCpds
        self.indexed_peaks, self.formula_tree, self.indexed_empCpds

        Future updates can include enforcing overlap in scans not just rtime tolerance.
        '''
        # check if intensity value is provided
        if 'representative_intensity' not in list_peaks[0]:
            if 'peak_area' in list_peaks[0]:
                for p in list_peaks:
                    p['representative_intensity'] = p['peak_area']
            elif 'height' in list_peaks[0]:
                for p in list_peaks:
                    p['representative_intensity'] = p['height']
            else:
                print("Warning: cannot get representative_intensity..")
                for p in list_peaks:
                    p['representative_intensity'] = 0

        self.list_peaks = list_peaks
        ECCON = epdsConstructor(list_peaks, mode=self.mode)
        self.dict_empCpds = ECCON.peaks_to_epdDict(
                        self.isotope_search_patterns,
                        self.adduct_patterns,
                        self.extended_adducts,
                        self.mz_tolerance_ppm,
                        self.rt_tolerance,
                        charges,
                        has_parent_masstrack,
        ) 
        self.index_empCpds()

    def build_from_list_empCpds(self, list_empCpds):
        for E in list_empCpds:
            self.dict_empCpds[E['interim_id']] = E
        self.index_empCpds()

    def index_empCpds(self):
        '''
        Build indices for self.list_peaks and self.empCpds.
        Khipu has peak:empCpd built N:1 but other methods should comply too.

        Updates
        -------
        self.indexed_peaks, self.formula_tree, self.indexed_empCpds
        '''
        for P in self.list_peaks:
            self.dict_peaks[P['id_number']] = P

        self.indexed_peaks  = build_centurion_tree(self.list_peaks)
        # use formula.get_formula_ions_tree
        self.formula_tree = get_formula_ions_tree(list_formula_mass, mode=self.mode)

        __PL = []
        for interim_id, epd in self.dict_empCpds.items():
            peaks = epd['MS1_pseudo_Spectra']
            for P in peaks:
                self.peak_to_empCpd[P['id_number']] = interim_id
                self.peak_to_empCpd_ion_relation[P['id_number']] = P['ion_relation']
                P['parent_epd_id'] = epd['interim_id'] 
                __PL.append( P )

        self.indexed_empCpds = build_centurion_tree(__PL)

    # standalone search functions
    def search_peaks_mz_single(self, query_mz, mz_tolerance_ppm=5):
        '''
        return list of matched peaks
        '''
        return find_all_matches_centurion_indexed_list(query_mz, self.indexed_peaks, mz_tolerance_ppm)

    def search_peaks_mz_batch(self, query_mz_list, mz_tolerance_ppm=5):
        results = []
        for query_mz in query_mz_list:
            results.append(
                self.search_peaks_mz_single(query_mz, mz_tolerance_ppm)
            )
        return results

    def search_empCpds_mz_single(self, query_mz, mz_tolerance_ppm=5):
        '''
        return list of matched empCpds
        '''
        return find_all_matches_centurion_indexed_list(query_mz, self.indexed_empCpds, mz_tolerance_ppm)

    def search_empCpds_mz_batch(self, query_mz_list, mz_tolerance_ppm=5):
        results = []
        for query_mz in query_mz_list:
            results.append(
                self.search_empCpds_mz_single(query_mz, mz_tolerance_ppm)
            )
        return results

    def search_peaks_compound_single(self, compound, mz_tolerance_ppm=5):
        '''
        Search a compound against this experimental dataset, and return matched peaks.
        Compound format:
        { 'primary_id': HMDB0000195, 'name': 'Inosine',
            "neutral_formula": C10H12N4O5,
            "neutral_formula_mass": 268.08077, ..}
        '''
        expected_peaks = compute_adducts_formulae(
            compound['neutral_formula_mass'], compound['neutral_formula'], self.mode, primary_only=False)
        # format [[304.203251, 'M[1+]', 'C19H28O3'], ...]
        return list(zip( [x[1] for x in expected_peaks],
                    self.search_peaks_mz_batch([x[0] for x in expected_peaks], mz_tolerance_ppm)
                    ))

    def search_peaks_compound_batch(self, list_cpd, mz_tolerance_ppm=5):
        results = []
        for cpd in list_cpd:
            results.append(
                self.search_peaks_compound_single(cpd, cpd, mz_tolerance_ppm)
            )
        return results

    def annotate_empCpds_against_KCD(self, KCD, mz_tolerance_ppm=5):
        '''
        Get all empCpd matches between this experimental dataset and a known compound database.
        KCD: knownCompoundDatabase instance. KCD.search_emp_cpd_single() searches only neutral mass or anchor ion. 
        Returns dictionary e.g.
                {(101, []),
                (102,
                [('C9H20NO2_174.149404', 'M[1+]', 2),
                ('C9H17NO_155.131014', 'M+H2O+H[1+]', 1)]),
                (103,
                [('C6H15N4O2_175.119501', 'M[1+]', 2),
                ('C6H14N4O2_174.111676', 'M+H[1+]', 2)]),
                (104, []), ...}
        '''
        resultDict = {}
        for epd in self.dict_empCpds.values():
            resultDict[epd['interim_id']] = KCD.search_emp_cpd_single(epd, self.mode, mz_tolerance_ppm )
        return resultDict

    def search_mz_for_formula(self, mz, mz_tolerance_ppm=5):
        '''
        return best matched formula or None, 
        format as {'mz': ion[0], 'neutral_formula': formula, 
                               'ion_relation': ion[1],
                               'neutral_formula_mass': mass}
        '''
        return search_mz_formula_tree(mz, self.formula_tree, limit_ppm=mz_tolerance_ppm)

    # Annotation functions
    def extend_empCpd_annotation(self, KCD):
        '''
        This searches KCD for the list of empCpds.
        With khipu, most empCpds have neutral mass, which is used for KCD search.
        If no neutral mass, an anchor ion is used for KCD search (KCD.search_emp_cpd_single).
        Returned list_matches have KCD empCpd identifiers but compound records need to pull out KCD later.
        '''
        epd_search_result_dict = self.annotate_empCpds_against_KCD(KCD, self.mz_tolerance_ppm)
        for interim_id, V in epd_search_result_dict.items():
            if V:
                self.dict_empCpds[interim_id]['list_matches'] = V

    def empCpds_formula_search(self, KCD):
        '''
        Update self.dict_empCpds for formulae first by KCD search then .data.formula_tree.
        KCD: knownCompoundDatabase instance.
        '''
        epd_search_result_dict = self.annotate_empCpds_against_KCD(KCD, self.mz_tolerance_ppm)
        for interim_id, V in epd_search_result_dict.items():
            if V:
                # update empCpd with formula matches
                self.dict_empCpds[interim_id]['list_matches'] = V
                #       expecting probem from khipu to this
                self.dict_empCpds[interim_id]['neutral_formula'] = \
                    KCD.mass_indexed_compounds[V[0][0]]['neutral_formula']
                self.dict_empCpds[interim_id]['neutral_formula_mass'] = \
                    KCD.mass_indexed_compounds[V[0][0]]['neutral_formula_mass']
            else:
                # first peak should be anchor
                anchor_mz = self.dict_empCpds[interim_id]['MS1_pseudo_Spectra'][0]['mz']
                F = self.search_mz_for_formula(anchor_mz, self.mz_tolerance_ppm)
                if F:
                    self.dict_empCpds[interim_id]['neutral_formula'] = F['neutral_formula']
                    self.dict_empCpds[interim_id]['neutral_formula_mass'] = F['neutral_formula_mass']

    def __extend_empCpds__(self):
        '''
        Extend empCpds by 
        additional isotopes/adducts, based on mass2chem.formula.compute_adducts_formulae

        Hold off; no need now from khipu results.
        '''
        peakList = [P for P in self.list_peaks if P['id_number'] not in self.peak_to_empCpd]
        peakTree = build_centurion_tree(peakList)
        for k,E in self.dict_empCpds.items():
            if E['neutral_formula']:
                anchor = E['MS1_pseudo_Spectra'][0]
                expected_ions = compute_adducts_formulae(E['neutral_formula_mass'], 
                                                E['neutral_formula'], self.mode, primary_only=False)
                                                # [(58.53894096677, 'M+2H[2+]', result_formula), ...,]
                for ion in expected_ions:
                    matches = find_all_matches_centurion_indexed_list(ion[0], peakTree, self.mz_tolerance_ppm)
                    for peak in matches:
                        if is_coeluted(anchor, peak, rt_tolerance=self.rt_tolerance):
                            peak['ion_relation'] = ion[1]
                            E['MS1_pseudo_Spectra'].append(peak)
                            self.peak_to_empCpd_ion_relation[peak['id_number']] = peak['ion_relation']
                            self.peak_to_empCpd[peak['id_number']] = E['interim_id']
                            # this may overwrite 1:N relationships

    def singleton_formula_search(self, KCD):
        '''
        Search singletons for formulae first by KCD search then .data.formula_tree.
        KCD: knownCompoundDatabase instance.

        Returns
        -------
        List of first matched KCD compound for each singleton in self.dict_peaks, e.g.
        [('F17904', {'interim_id': 15321, 'neutral_formula_mass': 916.606449, 'neutral_formula': 'C56H84O10',...}), ...]
        '''
        found = []
        singletons = [p for p in self.dict_peaks if p not in self.peak_to_empCpd.keys()]
        for p in singletons:
            _mz = self.dict_peaks[p]['mz']
            list_matches = KCD.search_mz_single(_mz, self.mode, self.mz_tolerance_ppm)
            # [{'mz': 130.017306555, 'parent_epd_id': 'C4H3FN2O2_130.017856', 'ion_relation': 'M[1+]'}]
            if list_matches:
                # take 1st match only here; will model better in future
                _epd = KCD.mass_indexed_compounds[list_matches[0]['parent_epd_id']]
                _epd['isotope'] = '13C/12C'
                _epd['ion_relation'] = _epd['modification'] = list_matches[0]['ion_relation']
                found.append((p, _epd))
            else:
                # formula search
                F = self.search_mz_for_formula(_mz, self.mz_tolerance_ppm)
                if F:
                    found.append((p, F))            # p is id_number

        return found
    

    def annotate_singleton_mummichog(self, KCD):
        '''
        This applies to KCD using a metabolic model as the sole database.

        Updates
        -------
        self.dict_empCpds : {id: empCpd, ...}
        '''
        singletons = [p for p in self.dict_peaks if p not in self.peak_to_empCpd.keys()]
        for p in singletons:
            peak = self.dict_peaks[p]
            interim_id = 'epd_' + peak['id_number']
            list_matches = KCD.search_mz_single(peak['mz'], self.mode, self.mz_tolerance_ppm)
            # [{'mz': 130.017306555, 'parent_epd_id': 'C4H3FN2O2_130.017856', 'ion_relation': 'M[1+]'}]
            if list_matches:
                self.dict_empCpds[interim_id] = {'interim_id': interim_id,
                    'MS1_pseudo_Spectra': [peak],
                    'list_matches': [(LL['parent_epd_id'], LL['ion_relation'], 1) for LL in list_matches],
                }


    def annotate_singletons(self, KCD):
        '''
        Search singletons for formulae first by KCD search then .data.formula_tree.
        Add new empCpds to self.dict_empCpds as new empCpds, and extend adduct search.
        Assign new identifiers as (10000 +number of dict_empCpds + numerical count).

        Updates
        -------
        self.dict_empCpds : {id: empCpd, ...}
        '''
        formula_to_peaks = {}
        found = self.singleton_formula_search(KCD)
        # multiple singletons can match to same formula as different ions, 
        # if ion calculation differs btw here and KCD
        for p,F in found:
            k = F['neutral_formula']
            if k in formula_to_peaks:
                formula_to_peaks[k].append((p, F))
            else:
                formula_to_peaks[k] = [(p, F)]

        found_peaks = set([x[0] for x in found])
        peakList = [P for P in self.list_peaks if P['id_number'] not in self.peak_to_empCpd and
                                                  P['id_number'] not in found_peaks]
        peakTree = build_centurion_tree(peakList)

        new_id_start = len(self.dict_empCpds) + 10000
        for formula, PP in formula_to_peaks.items():
            neutral_formula_mass = PP[0][1]['neutral_formula_mass']
            P1 = self.dict_peaks[PP[0][0]]
            tmp = [P1, ]
            for jj in range(1, len(PP)):
                _P = self.dict_peaks[PP[jj][0]]
                if is_coeluted(tmp[-1], _P, rt_tolerance=self.rt_tolerance):
                    tmp.append(_P)
                else:
                    # not coeluted, new empCpd
                    new_id_start += 1
                    self.dict_empCpds[new_id_start] = {'interim_id': str(new_id_start),
                            'neutral_formula_mass': neutral_formula_mass, 'neutral_formula': formula,
                            'MS1_pseudo_Spectra': self.__extend_peakList__(
                                formula, neutral_formula_mass, tmp, peakTree, self.mz_tolerance_ppm),
                    }
                    tmp = [_P, ]

            new_id_start += 1
            self.dict_empCpds[new_id_start] = {'interim_id': str(new_id_start),
                    'neutral_formula_mass': neutral_formula_mass, 'neutral_formula': formula,
                    'MS1_pseudo_Spectra': self.__extend_peakList__(
                                formula, neutral_formula_mass, tmp, peakTree, self.mz_tolerance_ppm),
            }


    def __extend_peakList__(self, neutral_formula, neutral_formula_mass, epd_peaks, peakTree,
            mz_tolerance_ppm=5):
        '''
        Extend epd_peaks by 
        additional isotopes/adducts, based on mass2chem.formula.compute_adducts_formulae, which 
        produces [(58.53894096677, 'M+2H[2+]', result_formula), ...,]
        '''
        new = []
        anchor = epd_peaks[0]
        for ion in compute_adducts_formulae(neutral_formula_mass, neutral_formula,
                                        self.mode, primary_only=False):
            matches = find_all_matches_centurion_indexed_list(ion[0], peakTree, mz_tolerance_ppm)
            for peak in matches:
                if is_coeluted(anchor, peak, rt_tolerance=self.rt_tolerance):
                    peak['ion_relation'] = ion[1]
                    new.append(peak)
        return epd_peaks + new


    def annotate_all_against_KCD(self, KCD, mz_tolerance_ppm=5):
        '''
        Not used now.

        Get matches of both empCpds and singleton peaks only in KCD.
        This function should not be used if one uses empCpds_formula_search and annotate_by_formula_grid.
        Results contain annotation via empCpd, via singleton, or empCpd features as if singleton.

        KCD: knownCompoundDatabase instance.
        return: peak_result_dict, epd_search_result_dict
        '''
        peak_result_dict = {}
        epd_search_result_dict = self.annotate_empCpds_against_KCD(KCD)
        for peak in self.list_peaks:
            ii, mz, apex = peak['id_number'], peak['mz'], peak['apex']
            matched_DB_shorts, matched_DB_records = '', ''
            interim_id = self.peak_to_empCpd.get(ii, None)  # id on expt data; ii could be 0
            epd_ion_relation = ''
            list_matches = []
            if interim_id != None:
                epd_ion_relation = self.peak_to_empCpd_ion_relation[ii]
                # Meaning there's empCpd for this feature, we look for earlier search result in search_result_dict
                list_matches = epd_search_result_dict[interim_id]
                # [('C7H8N2O_136.063663', 'M+H[1+]', 2), ('C7H9N2O_137.071488', 'M[1+]', 2), ('C7H6N2_118.053098', 'M+H2O+H[1+]', 1)]
                # this can be empty, but means the empCpd is not found in KCD.
            if not list_matches:
                # Meaning singleton feature OR unmatched empCpd feature, new search by m/z
                list_matches = KCD.search_mz_single(mz, self.mode, mz_tolerance_ppm)
                # [{'mz': 130.017306555, 'parent_epd_id': 'C4H3FN2O2_130.017856', 'ion_relation': 'M[1+]'}]
                list_matches = [(x['parent_epd_id'], x['ion_relation'], None) for x in list_matches]
            if list_matches:
                peak_result_dict[ii] = {'peak': peak,
                                    'interim_id': interim_id,               # this is empCpd id
                                    'epd_ion_relation': epd_ion_relation,
                                    'list_matches': list_matches,
                                    }

        return peak_result_dict, epd_search_result_dict


    def choose_top_epd(list_KCD_matches):
        '''recommend best from matches, e.g. 
        [('C6H14N_100.112624', 'M[1+]', 2), ('C6H13N_99.104799', 'M+H[1+]', 2)]
        by 1) highest score, and 2) peck order of ions (M+H > M+Na > M+ > M+H2O+H)
        '''
        pass


    def export_empCpds(self, outfile="EED_empCpds.json"):
        with open(outfile, 'w', encoding='utf-8') as O:
            json.dump(list(self.dict_empCpds.values()), O, ensure_ascii=False, indent=2)


    def export_annotations(self, KCD, export_file_name_prefix):
        '''
        Export expt empCpds and a tsv table for input list_peaks, with all matched db empCpds.
        '''
        self.export_empCpds(export_file_name_prefix+"EED_empCpds.json")
        s = "[peak]id_number\tmz\trtime\tapex(scan number)\t[EmpCpd]interim_id\t[EmpCpd]ion_relation\tneutral_formula\tneutral_formula_mass\
        \tname_1st_guess\tmatched_DB_shorts\tmatched_DB_records\n"
        
        for _, V in self.dict_empCpds.items():
            name_1st_guess, matched_DB_shorts, matched_DB_records = '', '', ''
            if 'list_matches' in V:
                list_matches = V['list_matches']
                if list_matches:
                    name_1st_guess = KCD.mass_indexed_compounds[list_matches[0][0]]['compounds'][0]['name']
                    matched_DB_shorts = ", ".join([ "(" + KCD.short_report_emp_cpd(xx[0]) + ")"  for xx in list_matches])
                    matched_DB_records = ", ".join([str(xx) for xx  in list_matches])

            for peak in V['MS1_pseudo_Spectra']:
                s += '\t'.join([str(x) for x in [
                    peak['id_number'], peak['mz'], peak['rtime'], peak['apex'], V['interim_id'], peak.get('ion_relation', ''),
                    V['neutral_formula'], V['neutral_formula_mass'],
                    name_1st_guess, matched_DB_shorts, matched_DB_records]]) + "\n"

        outfile = export_file_name_prefix + '.tsv'
        with open(outfile, encoding='utf-8', mode='w') as O:
            O.write(s)

        print("\nAnnotation of %d Empirical compounds was written to %s." %(len(self.dict_empCpds), outfile))



    #-----------------------------------------

    def update_annotation(self, KCD, peak_result_dict, adduct_patterns, mz_tolerance_ppm):
        '''
        Placeholder. Not used now but for future consieration -

        Update self.dict_empCpds, and extend search by formula-based additional isotopes/adducts.
        Not updating indexing or peak to empCpd mapping.

        peak_result_dict: result from self.annotate_all_against_KCD, which
         contains annotation via empCpd, via singleton, or empCpd features as if singleton.

        '''
        
        cpd_index_to_peaks = {}
        for p,v in peak_result_dict.items():
            cpd_index = v['list_matches'][0][0]
            if v['interim_id']:
                # peak has empCpd, update empCpd with formula matches
                self.dict_empCpds[v['interim_id']]['list_matches'] = v['list_matches']
                self.dict_empCpds[v['interim_id']]['neutral_formula'] = \
                    KCD.mass_indexed_compounds[cpd_index]['neutral_formula']
                self.dict_empCpds[v['interim_id']]['neutral_formula_mass'] = \
                    KCD.mass_indexed_compounds[cpd_index]['neutral_formula_mass']
            else:
                # compile to cpd_index_to_peaks, in order to group peaks by cpd_index e.g. `C6H15NO3_149.105193`
                if cpd_index in cpd_index_to_peaks:
                    cpd_index_to_peaks[cpd_index].append(p)
                else:
                    cpd_index_to_peaks[cpd_index] = [p]

        # create new empCpds based on cpd_index_to_peaks
        new_id_start = max(self.dict_empCpds.keys())
        for cpd_index, peaks in cpd_index_to_peaks.items():
            new_id_start += 1
            self.dict_empCpds[new_id_start] = {'interim_id': new_id_start,
                    'neutral_formula_mass': KCD.mass_indexed_compounds[cpd_index]['neutral_formula_mass'],
                    'neutral_formula': KCD.mass_indexed_compounds[cpd_index]['neutral_formula'],
                    'MS1_pseudo_Spectra': [self.dict_peaks[p] for p in peaks]
            }

        # try to assign remaining peaks to empCpds
        remaining_peaks = [p for p in self.dict_peaks if p not in 
                            set(list(peak_result_dict.keys()) + list(self.peak_to_empCpd.keys()))]

        ECCON = epdsConstructor(remaining_peaks, mode=self.mode)
        list_empCpds = ECCON.extend_empCpds_by_formula_grid(


            self.dict_empCpds.values(), remaining_peaks, adduct_patterns, mz_tolerance_ppm=mz_tolerance_ppm,
        ) 
        # update self.dict_empCpds
        for EC in list_empCpds:
            self.dict_empCpds[EC['interim_id']] = EC



"""
Function used to port different genome-scale metabolite models into metDataModel.

Yuanye Chi 2023-03-22
"""

from jms.port.port_config import *
from jms.port.port_CUT import port_CUT
from jms.port.port_AGORA import port_AGORA
    

def port(model_source):

    if model_source == Sources.CUT:
        for species in basic_info_dict[Sources.CUT].keys():
            port_CUT(species)
    elif model_source == Sources.AGORA:
        port_AGORA()
        


 


'''
Test coverage of a dataset or a model on pathways/ontologies.

Could exclude currency metabolites in future.

'''

from .modelConvert import convert_json_model, DataMeetModel
from .search import build_centurion_tree, find_all_matches_centurion_indexed_list
from .data.humangem_pathways import humangem
# from .data.smpdb_pathways import smpdb
from .empiricalCpds import *



def report_khipu_statistics(json_epd_file, snr=5, shape=0.9, natural_ratio_limit=0.5):
    '''
    Get numbers of khipus and singletons, and isopairs and good khipus.
    Good khipu = with isopair and M0 being a good feature.

    num_isopair_mtracks is based on num_good_khipus. 
    Thus the ratio num_good_khipus/num_isopair_mtracks indicates ave good isomers observed.
    
    This is for data without isotope labeling. 
    Planing to use function report_khipu_statistics_from_labeled_data for labeled data.
    
    Returns stats_dict, good_khipus.

    '''
    list_epds = list(json.load(open(json_epd_file)).values())
    num_features, num_good_features = 0, 0
    for epd in list_epds:
        for f in epd["MS1_pseudo_Spectra"]:
            num_features += 1
            f['snr'] = float(f['snr'])
            f['goodness_fitting'] = float(f['goodness_fitting'])
            f['peak_area'] = float(f['peak_area'])
            f['is_good_peak'] = check_good_peak(f, snr, shape)
            if f['is_good_peak']:
                num_good_features += 1
    
    khipus_isopairs, num_isopair_mtracks, good_khipus = get_isopairs_good_khipus(list_epds, natural_ratio_limit)
    num_empcpds = len(list_epds)
    num_singletons = count_singletons(list_epds)
    num_good_khipus = len(good_khipus)
    if num_isopair_mtracks:
        ratio_good_khipus_mtracks = num_good_khipus/num_isopair_mtracks
    else:
        ratio_good_khipus_mtracks = None
    
    stats_dict = {
        'num_features': num_features,
        'num_good_features': num_good_features,
        'num_empcpds': num_empcpds,
        'num_khipus': num_empcpds - num_singletons,
        'num_singletons': num_singletons,
        'num_khipus_isopairs': len(khipus_isopairs),
        'num_good_khipus': num_good_khipus,
        'num_isopair_mtracks': num_isopair_mtracks,
        'ratio_good_khipus_mtracks': ratio_good_khipus_mtracks
    }
    
    return stats_dict, good_khipus


def report_khipu_statistics_from_labeled_data(isotope_labeling_ratio=1.2):
    '''
    
    This analysis is often sample dependent, as control samples are usually included in the expt.
    
    see .empiricalCpds.filter_khipus_by_samples
    
    '''
    pass



def get_feature_stats_per_table(infile, snr=5, shape=0.9, sep='\t'):
    '''
    Get num_features, num_good_features from asari feature table.
    Alternatively, one can get stats from the empCpd JSON file.
    '''
    def _is_good_feature(line, snr, shape, sep):
        a = line.split(sep)
        _snr, _shape = float(a[7]), float(a[6])
        if _snr > snr and _shape > shape:
            return True
        else:
            return False
        
    lines = open(infile).readlines()
    num_features = len(lines) - 3
    num_good_features = len([line for line in lines[3:] if _is_good_feature(line, snr, shape, sep)])

    return num_features, num_good_features
    

def check_good_peak(peak, snr=5, shape=0.9, area=0):
    '''
    Decide if peak is good quality based on SNR and peak shape.
    Not using peak area by default value.
    '''
    if peak['snr'] > snr and peak['goodness_fitting'] > shape and peak['peak_area'] > area:
        return True
    else:
        return False


def report_pathway_coverage_from_gem(model, list_neutral_epds, outfile):
    '''
    This writes a report outfile and returns metabolic_pathways
    One can choose to use all neutral epds or only good epds.
    e.g. 
    stats_dict, good_khipus = report_khipu_statistics(json_epd_file, snr=5, shape=0.9, natural_ratio_limit=0.5)
    neutrals = get_neutrals( good_khipus )
    
    model is the JSON GEM model, e.g.
    model = json.load(open('metabolicModel_az_HumanGEM_20220302_noCompartmentalization.json'))
    model.keys()    
    dict_keys(['id', 'list_of_reactions', 'list_of_compounds', 'list_of_pathways', 'meta_data'])

    report_pathway_coverage modifies the pathways - thus pathways are reloaded here everytime.
    '''
    mcgmodel = convert_json_model(model)
    metabolic_pathways = report_pathway_coverage_from_neutrals( mcgmodel, list_neutral_epds)
    export_pathway_coverage_table(metabolic_pathways, outfile)
    return metabolic_pathways


def report_pathway_coverage_from_neutrals(mcg_model, list_neutral_epds, ppm=5):
    '''Returns a list of pathways with matched cpds to list_epds.

    list_epds : list of empirical compounds, but mz is neutral mass.
    mcg_model : metabolic model in mummichog style.

    '''
    mztree = build_centurion_tree(list_neutral_epds)
    for p in mcg_model['metabolic_pathways']:
        cpds = [mcg_model['Compounds'].get(x, None) for x in p['cpds']]
        cpds = [x for x in cpds if x]
        for c in cpds:
            c['mw'] = c.get('neutral_mono_mass', None) or c.get('neutral_formula_mass', None)
        
        MM = get_match(cpds, mztree, ppm)
        p['number_cpds'] = len(p['cpds'])   # likely greater than cpds
        p['number_matched_epds'] = len(MM)
        p['matched_epds'] = [(x['id'], x['name'], x['mw'], y) for x,y in MM]
        p['cpd_names'] = [mcg_model['dict_cpds_def'].get(x, x) for x in p['cpds']]

    return mcg_model['metabolic_pathways']



def get_pathwayCoverage_cpds_to_epds(list_epds, pathway_collection=humangem, ppm=5):
    '''
    Match list_epds to list_of_metabolites in full metabolic model (i.e. pathway_collection).
    list_epds : input list of empirical compounds. Users can apply filtering before this.
    pathway_collection : humangem = {
                                'version': 'https://github.com/SysBioChalmers/Human-GEM, retrieved 2022-02-09',
                                'dict_pathways': dict_pathways,
                                'list_of_metabolites': list_of_metabolites,
                            }
    Returns
    -------
    List of matches as [(cpd['id'], cpd['name'], matched_list of epds), ...]

    Examples
    --------
    list_of_metabolites[55]:
        {'id': 'MAM00067', 'mw': 1101.43877624496, 'name': '(2E)-tricosenoyl-CoA', 'pathways': ['Fatty acid elongation (odd-chain)']}
    list(humangem['dict_pathways'].items())[11]:
        ('group12',
        {'name': 'Beta oxidation of branched-chain fatty acids (mitochondrial)',
        'cpds': ['MAM00845',
        'MAM01802',
        ...,
        'MAM02040']})
    '''
    neutrals = get_neutrals(list_epds)
    mztree = build_centurion_tree(neutrals)
    matched = []
    for cpd in pathway_collection['list_of_metabolites']:
        MM = find_all_matches_centurion_indexed_list(cpd['mw'], mztree, ppm)
        if MM:
            matched.append((cpd['id'], cpd['name'], MM))

    return matched

def get_pathwayCoverage_mummichog(list_epds, metabolic_model, ppm=5):
    '''Match list_epds to pathways in mummichog style, using modelConvert.DataMeetModel.
    '''
    DMM = DataMeetModel(MetabolicModel=metabolic_model, userFeatureList=None, userListEmpCpds=list_epds)
    DMM.mz_tolerance_ppm = ppm
    return DMM.match_all()


#
# ------------------------------------------------------------------------------------------
# Convenience functions below for quick analysis; not suitabe for production software 

def report_pathway_coverage(mcg_model, list_epds, neutral_formula_mass=True, multiple_ions=True, ppm=5):
    '''Returns a list of pathways with matched cpds to list_epds.
    This test each pathway in mcg_model and collect matched epds.

    list_epds : list of empirical compounds.
    mcg_model : metabolic model in mummichog style.

    returns metabolic_pathways.
    '''
    neutrals = get_neutrals( 
                            filter_epds(list_epds, neutral_formula_mass, multiple_ions) 
                            )
    mztree = build_centurion_tree(neutrals)
    for p in mcg_model['metabolic_pathways']:
        cpds = [mcg_model['Compounds'].get(x, None) for x in p['cpds']]
        cpds = [x for x in cpds if x]
        for c in cpds:
            c['mw'] = c.get('neutral_mono_mass', None) or c.get('neutral_formula_mass', None)
        
        MM = get_match(cpds, mztree, ppm)
        p['number_cpds'] = len(p['cpds'])   # likely greater than cpds
        p['number_matched_epds'] = len(MM)
        p['matched_epds'] = [(x['id'], x['name'], x['mw'], y) for x,y in MM]
        p['cpd_names'] = [mcg_model['dict_cpds_def'].get(x, x) for x in p['cpds']]

    return mcg_model['metabolic_pathways']


def export_pathway_coverage_table(metabolic_pathways, outfile='pathway_matches.tsv'):
    '''Export stats table from metabolic_pathways, after running coverage analysis.
    '''
    s = 'pid\tname\tN_matched\tN_in_pathway\tcpds_in_pathway\tmatched_epds\n'
    for p in metabolic_pathways:
        s += '\t'.join([p['id'], p['name'], str(p['number_matched_epds']), str(p['number_cpds']),
                        str(p['matched_epds']), 
                        str(p['cpd_names']), 
                        ]) + '\n'
    with open(outfile, 'w') as O:
        O.write(s)


def wrapper_file2coverage(json_model, epd_json_file, 
                          neutral_formula_mass=True, multiple_ions=True, ppm=5,
                          outfile='pathway_matches.tsv'):
    '''
    There are functions handling empCpd filtering in report_pathway_coverage.
    They are preferred to be done explicitly outside this.
    See report_pathway_coverage_from_gem
    '''
    list_epds = load_epds_from_json(epd_json_file)
    mcg_model = convert_json_model(json_model)
    metabolic_pathways = report_pathway_coverage(
        mcg_model, list_epds, neutral_formula_mass, multiple_ions, ppm)
    export_pathway_coverage_table(metabolic_pathways, outfile)




#
# ------------ test ---------------
#

if __name__ == "__main__":
    
    model = json.load(open('/Users/lish/li.proj/ipsc_atlas/metabolicModel_az_HumanGEM_20220302_noCompartmentalization.json'))
      
    t = '/Users/lish/li.workspace/MT_AML_13C/mzML_HILICneg_20240318_THP1_MOLM13_asari_project_32693656/Annotated_empiricalCompounds.json'
    stats_dict, good_khipus = report_khipu_statistics(t, snr=5, shape=0.9, natural_ratio_limit=0.5)
    neutrals = get_neutrals( good_khipus )
    print("\n\n~~~~~~~~~ num good neutrals ~~~~~~~~~~\n\n")
    print(len(neutrals))
    print(stats_dict)
    
    report_pathway_coverage_from_gem(model, neutrals, outfile='test_jms_coverage.tsv')
    


"""
This module implements the report object which is a wrapper around an fpdf object. 

The reports are meant to be a high-level overview of an experiment, the feature tables,
empcpds, etc and summarize some of the qaqc results.
"""

import os
import platform
import sys
import uuid
import datetime
import textwrap
from pip._vendor import pkg_resources
from fpdf import FPDF
import matplotlib.pyplot as plt
from . import FeatureTable
from . import utils

class ReportPDF(FPDF):
    """
    This class is simply a wrapper around FPDF from fpdf library so
    that we can have a consistent header and footer for all pages

    :param header_text: text to be included on the header of each page
    """
    def __init__(self, header_text):
        self.header_text = header_text
        super().__init__()

    def header(self):
        """
        This function is called on every page of the report. It 
        generates the header for the page. 
        """
        self.set_font('Arial', 'B', 15)
        self.cell(80)
        self.cell(30, 10, self.header_text, 0, 0, 'C')
        self.ln(10)

    def footer(self):
        """
        This function is called on every page of the report. It 
        generates the footer. Currently this does nothing but it 
        is required. 
        """
        self.set_y(15)
        self.set_font('Arial', 'I', 8)

class Report():
    """
    The report object allows the creation of a pdf report. The reports 
    are defined using templates. 

    A template is a JSON-formatted list of dictionaries. Each dictionary 
    corresponds one-to-one with a method of this object. Which method
    the dictionary is intended to call is defined by the dictionary's
    "section" field. The section field value should be the name of a 
    method for this object. 
    """
    def __init__(self, experiment, parameters) -> None:
        self.experiment = experiment
        self.parameters = parameters

        self.default_font = ['Arial', '', 12]
        report_title = 'PCPFM Report - ' + self.experiment.experiment_directory.split("/")[-1]
        self.report = ReportPDF(report_title)
        self.report.add_page()
        self.report.set_font(*self.default_font)
        self.max_width = round(self.report.line_width * 900,0)
        self.style = self.__preprocess_style(self.parameters["report_config"])
        self.__create_report()

    def __preprocess_style(self, style):
        """
        This function takes the provided style and checks that it has 
        the required save section and that the defined sections are 
        valid. Invalid sections are removed here. Furthermore, fields
        such as "texts" that can occur at the same level as "sections" 
        are then inserted into the approprate location. 

        :param style: the JSON configuration of the report

        :return: the style with text added and invalid methods deleted
        """
        for section in style["sections"]:
            for top_level_field in style.keys():
                if top_level_field in section:
                    if section[top_level_field] in style[top_level_field]:
                        section[top_level_field] = style[top_level_field][section[top_level_field]]
        valid_sections = []
        for section in style["sections"]:
            try:
                getattr(self, section["section"])
                valid_sections.append(section)
            except AttributeError:
                print(section["section"] + "is not a valid section!\nValid sections include: ")
                for method in dir(Report):
                    if not method.startswith("__") and not method.startswith("_Report__"):
                        print("\t", method)
        section_names = [x["section"] for x in valid_sections]
        if "save" not in section_names:
            print("Save section not found in report! will abort")
            sys.exit()

        expanded_sections_ft = []
        all_feature_tables = list(self.experiment.feature_tables.keys())
        for section in valid_sections:
            table_moniker = section.get("table", None)
            if table_moniker == "*":
                for feature_table in all_feature_tables:
                    new_section = dict(section)
                    new_section["table"] = feature_table
                    expanded_sections_ft.append(new_section)
            elif isinstance(table_moniker, list):
                for feature_table in table_moniker:
                    new_section = dict(section)
                    new_section["table"] = feature_table
                    expanded_sections_ft.append(new_section)
            else:
                expanded_sections_ft.append(dict(section))

        expanded_sections_ft_empcpd = []
        all_empcpds = list(self.experiment.empCpds.keys())
        for section in expanded_sections_ft:
            empcpd_moniker = section.get("empcpd", None)
            if empcpd_moniker == "*":
                for empcpd in all_empcpds:
                    new_section = dict(section)
                    new_section["empcpd"] = empcpd
                    expanded_sections_ft_empcpd.append(new_section)
            else:
                expanded_sections_ft_empcpd.append(dict(section))
        return expanded_sections_ft_empcpd

    def __create_report(self):
        """
        This function iterates through valid sections and generates the
        report element each section specifies.

        The way this works is that the section field should have a value whose name is the same
        as a method of the object, it finds and executes that method. As a result, every method
        must have the same call signature which results in some undesirable warnings from the IDE
        but is otherwise fine. This can be fixed in the future using something like the inspect
        module to see what the function requires and only pass that data.
        """
        for section in self.style:
            method = getattr(self, section["section"])
            try:
                if method:
                    if "table" in section:
                        if section["table"] + "_cleaned" in self.experiment.feature_tables:
                            section["table"] = section["table"] + "_cleaned"
                            method(section)
                        else:
                            method(section)
                    else:
                        method(section)
            except:
                pass
        self.experiment.save()


    def __reset_font(self):
        """
        This method resets the font of the report to the default. This 
        is needed because the report's font is set at a report-wide 
        level and must be changed each time a different font, size, or
        typesetting is required. 
        """
        self.report.set_font(*self.default_font)

    def __section_head(self, title):
        """
        This writes the header for a section. This uses a larger font
        and sets it to be in bold.
        
        :param title: the title of the section
        """
        self.report.cell(80)
        self.report.set_font(self.default_font[0], 'B', self.default_font[2])
        self.report.cell(30, 10, title, 0, 0, 'C')
        self.__reset_font()
        self.report.ln(5)

    def __section_line(self, content, options=None):
        """
        This writes a line in a section. Optionally, the font can be 
        made bold by passing 'B' in options. 

        TODO: all of these functions for writing lines can be 
        unified with a single command that takes multiple params

        :param content: what to write in the line
        :param options: an iterable with text options valid options
            include: 'B' for bold.
        """
        options = set(options) if options is not None else set()
        if "bold" in options:
            self.report.set_font(self.default_font[0], 'B', self.default_font[2])
            self.report.cell(30, 10, content, 0, 0, "B")
            self.__reset_font()
        else:
            self.report.cell(30, 10, content, 0, 0)
        self.report.ln(5)

    def __section_text(self, text, options=None):
        """
        This writes a block of text in a section, one line at a time
        using the __section_line function. This also handles chopping 
        up a line into pieces that can fit on the page (i.e., word 
        wrapping)

        TODO: all of these functions for writing lines can be 
        unified with a single command that takes multiple params

        :param text: what to write in the line
        :param options: an iterable with text options valid options
            include: 'B' for bold.
        """
        text = ' '.join(text.split(None))
        text = ' '.join(text.split("\n"))
        i = 0
        for line in textwrap.wrap(text, width=95):
            self.__section_line(line, options=options)
        self.report.ln(5)
        self.report.ln(5)

    def TICs(self, section_desc):
        """
        This generates, if not pre-existing, and includes the TIC of 
        each acquisition in the experiment to the report. 

        Requires: None
        """
        for acquisition in self.experiment.acquisitions:
            try:
                tic_path = acquisition.TIC()
                self.report.image(tic_path, w=self.max_width)
            except:
                print(section_desc)

    def experiment_summary(self, section_desc):
        """
        This will list the empcpds and feature tables in the experiment.

        Requires: None
        """ 
        self.__section_head("Experiment Summary")
        if 'text' in section_desc:
            self.__section_text(section_desc['text'])
        self.__section_line("empCpd list", options=["bold"])
        for empcpd_moniker in self.experiment.empCpds.keys():
            self.__section_line(empcpd_moniker)
        self.__section_line(' ')
        self.__section_line("Feature Table list", options=["bold"])
        for feature_table_moniker in self.experiment.feature_tables.keys():
            self.__section_line(feature_table_moniker)


    def annotation_summary(self, section_desc):
        """
        This summarizes the annotations for each empcpd in the experiment.

        This counts the number of annotations per empcpd at each level.

        Args:
            Requires: None
        """     
        self.__section_head("Annotation Summary")
        if 'text' in section_desc:
            self.__section_text(section_desc['text'])
        self.__section_line("Name, #EmpCpds, #l4 Annotated, #l2 Annotated, #l1b Annotated, #l1a Annotated", options=["bold"])
        for empcpd_moniker in self.experiment.empCpds.keys():
            empcpds = self.experiment.retrieve_empCpds(empcpd_moniker, True)
            num_l4_annotated = 0
            num_l2_annotated = 0
            num_l1b_annotated = 0
            num_l1a_annotated = 0
            for kp in empcpds.dict_empcpds.values():
                num_l1b_annotated += int(bool(kp.get("Level_1b", [])))
                num_l4_annotated += int(bool(kp.get("Level_4")))
                has_l2 = False
                has_l1a = False
                for ms2_spectrum in kp.get("MS2_Spectra", []):
                    for annotation in ms2_spectrum.get("annotations", []):
                        if annotation["annotation_level"] == "Level_2":
                            has_l2 = True
                        elif annotation["annotation_level"] == "Level_1a":
                            has_l1a = True
            num_l2_annotated += int(has_l2)
            num_l1a_annotated += int(has_l1a)
            line = ", ".join([str(x) for x in [empcpd_moniker, 
                                               str(len(empcpds.dict_empcpds)), 
                                               num_l4_annotated, 
                                               num_l2_annotated, 
                                               num_l1b_annotated, 
                                               num_l1a_annotated]])
            self.__section_line(line)

    def table_summary(self, section_desc):
        """
        This summarizes the feature tables in the experiment. 

        Requires: None
        """

        self.__section_head("Feature Table Summary")
        if 'text' in section_desc:
            self.__section_text(section_desc['text'])
        self.__section_line("Table Name, Num Samples, Num Features", options=["bold"])
        tables = [x for x in self.experiment.feature_tables.keys() if "cleaned" not in x]
        for table in tables:
            ft = self.experiment.retrieve_feature_table(table, True)
            line = ", ".join([str(x) for x in [table, ft.num_samples, ft.num_features]])
            self.__section_line(line)

    def empcpd_summary(self, section_desc):
        """
        This summarizes the feature tables in the experiment. 

        Requires: None
        """
        self.__section_head("empCpd Table Summary")
        if 'text' in section_desc:
            self.__section_text(section_desc['text'])
        self.__section_line("EmpCpd Name, Num Khipus, Num Features", options=["bold"])
        for empcpd in self.experiment.empCpds.keys():
            empcpd_object = self.experiment.retrieve_empCpds(empcpd, True)
            self.__section_line(", ".join([str(x) for x in [empcpd, empcpd_object.num_khipus, empcpd_object.num_features]]))

    def command_history(self, section_desc):
        """
        This summarizes each command that has been executed in the 
        analysis. 

        Requires: None
        """

        self.__section_head("Command History")
        for command in self.experiment.command_history:
            self.__section_text(command)

    def version_summary(self, section_desc):
        """
        This summarizes each command that has been executed in the 
        analysis. 

        Requires: None
        """
        self.__section_head("Software Version Summary")
        self.__section_line(":".join(['pcpfm', pkg_resources.get_distribution('pcpfm').version]))
        _package = pkg_resources.working_set.by_key['pcpfm']
        for req in _package.requires():
            version = pkg_resources.get_distribution(req.name).version
            self.__section_line(":".join([req.name, version]))
        self.__section_line('')
        self.__section_line("OS: " + platform.system())
        self.__section_line("Python Version: " + platform.python_version())
        self.__section_line("Architecture: " + platform.machine())
        self.__section_text("Uname: " + " ".join(platform.uname()))

    def computational_performance(self, section_desc):
        """
        This summarizes each command and computes the time required
        for that step. This is useful for benchmarking.

        Requires: None
        """

        self.__section_head("Computational Performance")
        tn_minus_one = None
        current_time = None
        command_order, time_required = [], []
        start_time = None
        for command in self.experiment.command_history:
            tn_minus_one = current_time
            current_time = float(command.split(":")[0])
            if command.endswith("start_analysis"):
                start_time = float(current_time)
            else:
                time_sums = (current_time - tn_minus_one) / 60
                command_order.append(command)
                time_required.append(time_sums)
        total_time = current_time - start_time
        self.__section_text("Total time for analysis: " + str(total_time / 60) + " minutes")
        plt.bar(command_order, time_required)
        plt.title("Time Required per Command")
        name = "/tmp/" + str(uuid.uuid4) + ".png"
        plt.savefig(name)
        self.report.image(name)

    def timestamp(self, section_desc):
        """
        This will timestamp the report.

        Requires: None
        """
        timestamp_string = 'Report generated on ' + str(datetime.datetime.now())
        self.__section_head("Timestamp")
        self.__section_line(timestamp_string)

    def save(self, section_desc):
        """
        This saves the report pdf to the specified location

        Requires: "report_name"
        """
        output_subdir = os.path.abspath(self.experiment.output_subdirectory)
        report_path = os.path.join(output_subdir, section_desc["report_name"])
        if not report_path.endswith(".pdf"):
            report_path += ".pdf"
        print("saving: ", report_path)
        self.report.output(report_path)

    def figure(self, section_desc):
        """
        This inserts a figure into the report. Since all figures are 
        currently QAQC figures from feature tables, this method 
        requires specifying a table name and a qaqc result to 
        visualize. 

        Requires: "table" - the moniker for the table
                  "name" - name for the qaqc result to add figure of

        The valid fields of name are populated at runtime for this 
        method and will be displayed if an incorrect name field is 
        provided.

        """
        feature_table = self.experiment.retrieve_feature_table(section_desc["table"], True)
        params_for_figure = dict(self.parameters)
        params_for_figure['all'] = False
        params_for_figure['save_plots'] = True
        feature_table.generate_figure_params(params_for_figure)
        figure_path = feature_table.save_fig_path(section_desc["name"])
        try:
            if os.path.exists(figure_path):
                self.report.add_page()
                self.__section_line("Table: " + section_desc["table"] + "  " + "Figure: " + section_desc["name"])
                self.report.ln(10)
                self.report.image(figure_path, w=self.max_width-10)
            else:
                feature_table = self.experiment.retrieve_feature_table(section_desc["table"], True)
                params_for_figure = dict(self.parameters)
                params_for_figure['all'] = False
                params_for_figure['save_plots'] = True
                if section_desc["name"] in feature_table.qaqc_result_to_key:
                    params_for_figure[feature_table.qaqc_result_to_key[section_desc["name"]]] = True
                    feature_table.QAQC(params_for_figure)
                    self.report.add_page()
                    self.__section_line("Table: " + section_desc["table"] + "  " + "Figure: " + section_desc["name"])
                    self.report.ln(10)
                    self.report.image(figure_path, w=self.max_width)
            if "text" in section_desc and section_desc["text"]:
                self.__section_text(section_desc["text"])
        except:
            pass

# this updates the docstring for report
qaqc_names = list(FeatureTable.FeatureTable.qaqc_result_to_key.keys())
getattr(Report, "figure").__doc__ += "\nValid name values are: \n\t" + "\n\t".join(qaqc_names)
